// Class: Minnow
//
// Author: Alyce Brady
//
// This class is based on the College Board's Fish class,
// as allowed by the GNU General Public License.  Fish
// is a class within the AP(r) CS Marine Biology Simulation
// case study (see www.collegeboard.com/ap/students/compsci).
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

import java.awt.Color;
import java.util.ArrayList;
import java.util.Random;

/**
 *  Minnow Program:<br>
 *
 *  The Minnow class represents a simple fish in an environment.  It is
 *  similar to, but simpler than, the Fish class in the Marine Biology
 *  Simulation Case Study.
 *
 *  <p>
 *  When moving on to the Marine Biology Simulation case study, notice the
 *  following differences between the Minnow and Fish classes:
 *  <pre>
 *      The Fish class has more instance variables and more constructors.
 *          (see pp. 27 - 30 of the MBS case study narrative)
 *      Fish check that they are still in the environment before acting.
 *          (p. 33)
 *      Fish methods use the changeDirection helper method and the
 *          environment, location, and direction accessor methods.  (p. 39)
 *      The Fish helper methods are protected, not private.  (pp. 30, 65)
 *  </pre>
 *
 *  @author Alyce Brady
 *  @version 1 September 2002
 *  @see Environment
 *  @see Direction
 *  @see Location
 **/

public class Minnow implements Locatable
{
    // Instance Variables: Encapsulated data for EACH minnow
    private Environment theEnv;        // environment in which the minnow lives
    private Location myLoc;            // minnow's location
    private Direction myDir;           // minnow's direction
    private Color myColor;             // minnow's color


  // constructor and related helper method

    /** Constructs a minnow at the specified location in a given environment.
     *  The Minnow is assigned a random direction and random color.
     *  (Precondition: parameters are non-null; <code>loc</code> is valid
     *  for <code>env</code>.)
     *  @param env    environment in which minnow will live
     *  @param loc    location of the new minnow in <code>env</code>
     **/
    public Minnow(Environment env, Location loc)
    {
        theEnv = env;
        myLoc = loc;
        myDir = env.randomDirection();
        myColor = randomColor();
        theEnv.add(this);    // object is now at location myLoc in environment
    }

    /** Generates a random color.
     *  @return       the new random color
     **/
    private Color randomColor()
    {
        // There are 256 possibilities for the red, green, and blue attributes
        // of a color.  Generate random values for each color attribute.
        Random randNumGen = RandNumGenerator.getInstance();
        return new Color(randNumGen.nextInt(256),    // amount of red
                         randNumGen.nextInt(256),    // amount of green
                         randNumGen.nextInt(256));   // amount of blue
    }


  // accessor methods

    /** Returns this minnow's environment.
     *  @return        the environment in which this minnow lives
     **/
    public Environment environment()
    {
        return theEnv;
    }

    /** Returns this minnow's color.
     *  @return        the color of this minnow
     **/
    public Color color()
    {
        return myColor;
    }

    /** Returns this minnow's location.
     *  @return        the location of this minnow in the environment
     **/
    public Location location()
    {
        return myLoc;
    }

    /** Returns this minnow's direction.
     *  @return        the direction in which this minnow is facing
     **/
    public Direction direction()
    {
        return myDir;
    }

    /** Returns a string representing key information about this minnow.
     *  @return  a string indicating the minnow's location and direction
     **/
    public String toString()
    {
        return "M" + myLoc.toString() + myDir.toString();
    }


  // modifier method

    /** Acts for one step in the simulation.
     **/
    public void act()
    {
        // A simulation asks objects to "act" -- to do whatever that
        // object should do in a single simulation timestep.  A minnow's
        // action is simply to move.
        move();
    }


  // internal helper methods

    /** Moves this minnow in its environment.
     **/
    private void move()
    {
        // Find a location to move to.
        Debug.print("Minnow " + toString() + " attempting to move.  ");
        Location nextLoc = nextLocation();

        // If the next location is different, move there.
        if ( ! nextLoc.equals(myLoc) )
        {
            // Move to new location.
            Location oldLoc = myLoc;
            changeLocation(nextLoc);

            // Update direction in case minnow had to turn to move.
            myDir = theEnv.getDirection(oldLoc, nextLoc);
            Debug.println("  Moves to " + myLoc + myDir);
        }
        else
            Debug.println("  Does not move.");
    }

    /** Helper Method -- Finds this minnow's next location.
     *  @return    the next location for this minnow
     **/
    private Location nextLocation()
    {
        // Return the location in front if it's empty.
        Location forwardLoc = theEnv.getNeighbor(myLoc, myDir);
        if ( theEnv.isEmpty(forwardLoc) )
            return forwardLoc;
        else
            return myLoc;
    }

    /** Finds empty locations adjacent to this minnow.
     *  @return    an ArrayList containing neighboring empty locations
     **/
    private ArrayList emptyNeighbors()
    {
        // Get all the neighbors of this minnow, empty or not.
        ArrayList nbrs = environment().neighborsOf(location());

        // Figure out which neighbors are empty and add those to a new list.
        ArrayList emptyNbrs = new ArrayList();
        for ( int index = 0; index < nbrs.size(); index++ )
        {
            Location loc = (Location) nbrs.get(index);
            if ( environment().isEmpty(loc) )
                emptyNbrs.add(loc);
        }

        return emptyNbrs;
    }

    /** Modifies this minnow's location and notifies the environment.
     *  @param  newLoc    new location value
     **/
    private void changeLocation(Location newLoc)
    {
        // Change location and notify the environment.
        Location oldLoc = location();
        myLoc = newLoc;
        environment().recordMove(this, oldLoc);

        // object is again at location myLoc in environment
    }

}
